import sys
import struct
import binascii

import _msbt as bt

bt.initwinsock()

L2CAP=0
RFCOMM=3

PORT_ANY=0

# Service class identifiers of standard services and service groups
SDP_SERVER_CLASS = "1000"
BROWSE_GRP_DESC_CLASS = "1001"
PUBLIC_BROWSE_GROUP = "1002"
SERIAL_PORT_CLASS = "1101"
LAN_ACCESS_CLASS = "1102"
DIALUP_NET_CLASS = "1103"
IRMC_SYNC_CLASS = "1104"
OBEX_OBJPUSH_CLASS = "1105"
OBEX_FILETRANS_CLASS = "1106"
IRMC_SYNC_CMD_CLASS = "1107"
HEADSET_CLASS = "1108"
CORDLESS_TELEPHONY_CLASS = "1109"
AUDIO_SOURCE_CLASS = "110a"
AUDIO_SINK_CLASS = "110b"
AV_REMOTE_TARGET_CLASS = "110c"
ADVANCED_AUDIO_CLASS = "110d"
AV_REMOTE_CLASS = "110e"
VIDEO_CONF_CLASS = "110f"
INTERCOM_CLASS = "1110"
FAX_CLASS = "1111"
HEADSET_AGW_CLASS = "1112"
WAP_CLASS = "1113"
WAP_CLIENT_CLASS = "1114"
PANU_CLASS = "1115"
NAP_CLASS = "1116"
GN_CLASS = "1117"
DIRECT_PRINTING_CLASS = "1118"
REFERENCE_PRINTING_CLASS = "1119"
IMAGING_CLASS = "111a"
IMAGING_RESPONDER_CLASS = "111b"
IMAGING_ARCHIVE_CLASS = "111c"
IMAGING_REFOBJS_CLASS = "111d"
HANDSFREE_CLASS = "111e"
HANDSFREE_AGW_CLASS = "111f"
DIRECT_PRT_REFOBJS_CLASS = "1120"
REFLECTED_UI_CLASS = "1121"
BASIC_PRINTING_CLASS = "1122"
PRINTING_STATUS_CLASS = "1123"
HID_CLASS = "1124"
HCR_CLASS = "1125"
HCR_PRINT_CLASS = "1126"
HCR_SCAN_CLASS = "1127"
CIP_CLASS = "1128"
VIDEO_CONF_GW_CLASS = "1129"
UDI_MT_CLASS = "112a"
UDI_TA_CLASS = "112b"
AV_CLASS = "112c"
SAP_CLASS = "112d"
PNP_INFO_CLASS = "1200"
GENERIC_NETWORKING_CLASS = "1201"
GENERIC_FILETRANS_CLASS = "1202"
GENERIC_AUDIO_CLASS = "1203"
GENERIC_TELEPHONY_CLASS = "1204"
UPNP_CLASS = "1205"
UPNP_IP_CLASS = "1206"
UPNP_PAN_CLASS = "1300"
UPNP_LAP_CLASS = "1301"
UPNP_L2CAP_CLASS = "1302"
VIDEO_SOURCE_CLASS = "1303"
VIDEO_SINK_CLASS = "1304"

# standard profile descriptor identifiers

SDP_SERVER_PROFILE = ( SDP_SERVER_CLASS, 0x0100 )
BROWSE_GRP_DESC_PROFILE = ( BROWSE_GRP_DESC_CLASS, 0x0100 )
SERIAL_PORT_PROFILE = ( SERIAL_PORT_CLASS, 0x0100 )
LAN_ACCESS_PROFILE = ( LAN_ACCESS_CLASS, 0x0100 )
DIALUP_NET_PROFILE = ( DIALUP_NET_CLASS, 0x0100 )
IRMC_SYNC_PROFILE = ( IRMC_SYNC_CLASS, 0x0100 )
OBEX_OBJPUSH_PROFILE = ( OBEX_OBJPUSH_CLASS, 0x0100 )
OBEX_FILETRANS_PROFILE = ( OBEX_FILETRANS_CLASS, 0x0100 )
IRMC_SYNC_CMD_PROFILE = ( IRMC_SYNC_CMD_CLASS, 0x0100 )
HEADSET_PROFILE = ( HEADSET_CLASS, 0x0100 )
CORDLESS_TELEPHONY_PROFILE = ( CORDLESS_TELEPHONY_CLASS, 0x0100 )
AUDIO_SOURCE_PROFILE = ( AUDIO_SOURCE_CLASS, 0x0100 )
AUDIO_SINK_PROFILE = ( AUDIO_SINK_CLASS, 0x0100 )
AV_REMOTE_TARGET_PROFILE = ( AV_REMOTE_TARGET_CLASS, 0x0100 )
ADVANCED_AUDIO_PROFILE = ( ADVANCED_AUDIO_CLASS, 0x0100 )
AV_REMOTE_PROFILE = ( AV_REMOTE_CLASS, 0x0100 )
VIDEO_CONF_PROFILE = ( VIDEO_CONF_CLASS, 0x0100 )
INTERCOM_PROFILE = ( INTERCOM_CLASS, 0x0100 )
FAX_PROFILE = ( FAX_CLASS, 0x0100 )
HEADSET_AGW_PROFILE = ( HEADSET_AGW_CLASS, 0x0100 )
WAP_PROFILE = ( WAP_CLASS, 0x0100 )
WAP_CLIENT_PROFILE = ( WAP_CLIENT_CLASS, 0x0100 )
PANU_PROFILE = ( PANU_CLASS, 0x0100 )
NAP_PROFILE = ( NAP_CLASS, 0x0100 )
GN_PROFILE = ( GN_CLASS, 0x0100 )
DIRECT_PRINTING_PROFILE = ( DIRECT_PRINTING_CLASS, 0x0100 )
REFERENCE_PRINTING_PROFILE = ( REFERENCE_PRINTING_CLASS, 0x0100 )
IMAGING_PROFILE = ( IMAGING_CLASS, 0x0100 )
IMAGING_RESPONDER_PROFILE = ( IMAGING_RESPONDER_CLASS, 0x0100 )
IMAGING_ARCHIVE_PROFILE = ( IMAGING_ARCHIVE_CLASS, 0x0100 )
IMAGING_REFOBJS_PROFILE = ( IMAGING_REFOBJS_CLASS, 0x0100 )
HANDSFREE_PROFILE = ( HANDSFREE_CLASS, 0x0100 )
HANDSFREE_AGW_PROFILE = ( HANDSFREE_AGW_CLASS, 0x0100 )
DIRECT_PRT_REFOBJS_PROFILE = ( DIRECT_PRT_REFOBJS_CLASS, 0x0100 )
REFLECTED_UI_PROFILE = ( REFLECTED_UI_CLASS, 0x0100 )
BASIC_PRINTING_PROFILE = ( BASIC_PRINTING_CLASS, 0x0100 )
PRINTING_STATUS_PROFILE = ( PRINTING_STATUS_CLASS, 0x0100 )
HID_PROFILE = ( HID_CLASS, 0x0100 )
HCR_PROFILE = ( HCR_SCAN_CLASS, 0x0100 )
HCR_PRINT_PROFILE = ( HCR_PRINT_CLASS, 0x0100 )
HCR_SCAN_PROFILE = ( HCR_SCAN_CLASS, 0x0100 )
CIP_PROFILE = ( CIP_CLASS, 0x0100 )
VIDEO_CONF_GW_PROFILE = ( VIDEO_CONF_GW_CLASS, 0x0100 )
UDI_MT_PROFILE = ( UDI_MT_CLASS, 0x0100 )
UDI_TA_PROFILE = ( UDI_TA_CLASS, 0x0100 )
AV_PROFILE = ( AV_CLASS, 0x0100 )
SAP_PROFILE = ( SAP_CLASS, 0x0100 )
PNP_INFO_PROFILE = ( PNP_INFO_CLASS, 0x0100 )
GENERIC_NETWORKING_PROFILE = ( GENERIC_NETWORKING_CLASS, 0x0100 )
GENERIC_FILETRANS_PROFILE = ( GENERIC_FILETRANS_CLASS, 0x0100 )
GENERIC_AUDIO_PROFILE = ( GENERIC_AUDIO_CLASS, 0x0100 )
GENERIC_TELEPHONY_PROFILE = ( GENERIC_TELEPHONY_CLASS, 0x0100 )
UPNP_PROFILE = ( UPNP_CLASS, 0x0100 )
UPNP_IP_PROFILE = ( UPNP_IP_CLASS, 0x0100 )
UPNP_PAN_PROFILE = ( UPNP_PAN_CLASS, 0x0100 )
UPNP_LAP_PROFILE = ( UPNP_LAP_CLASS, 0x0100 )
UPNP_L2CAP_PROFILE = ( UPNP_L2CAP_CLASS, 0x0100 )
VIDEO_SOURCE_PROFILE = ( VIDEO_SOURCE_CLASS, 0x0100 )
VIDEO_SINK_PROFILE = ( VIDEO_SINK_CLASS, 0x0100 )

SERVICE_RECORD_HANDLE_ATTRID = 0x0000
SERVICE_CLASS_ID_LIST_ATTRID = 0x0001
SERVICE_RECORD_STATE_ATTRID = 0x0002
SERVICE_ID_ATTRID = 0x0003
PROTOCOL_DESCRIPTOR_LIST_ATTRID = 0x0004
BROWSE_GROUP_LIST_ATTRID = 0x0005
LANGUAGE_BASE_ATTRID_LIST_ATTRID = 0x0006
SERVICE_INFO_TIME_TO_LIVE_ATTRID = 0x0007
SERVICE_AVAILABILITY_ATTRID = 0x0008
BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRID = 0x0009
DOCUMENTATION_URL_ATTRID = 0x000a
CLIENT_EXECUTABLE_URL_ATTRID = 0x000b
ICON_URL_ATTRID = 0x000c
SERVICE_NAME_ATTRID = 0x0100
SERVICE_DESCRIPTION_ATTRID = 0x0101
PROVIDER_NAME_ATTRID = 0x0102


def is_valid_address( s ):
    """
    returns True if address is a valid Bluetooth address

    valid address are always strings of the form XX:XX:XX:XX:XX:XX
    where X is a hexadecimal character.  For example,
        01:23:45:67:89:AB is a valid address, but
        IN:VA:LI:DA:DD:RE is not
    """
    try:
        pairs = s.split(":")
        if len(pairs) != 6: return False
        for b in pairs: int(b, 16)
    except:
        return False
    return True

def is_valid_uuid(uuid):
    """
    is_valid_uuid(uuid) -> bool

    returns True if uuid is a valid 128-bit UUID.

    valid UUIDs are always strings taking one of the following forms:
        XXXX
        XXXXXXXX
        XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
    where each X is a hexadecimal digit (case insensitive)
    """
    try:
        if len(uuid) == 4:
            if int( uuid, 16 ) < 0: return False
        elif len(uuid) == 8:
            if int( uuid, 16 ) < 0: return False
        elif len(uuid) == 36:
            pieces = uuid.split("-")
            if len(pieces) != 5 or \
                    len(pieces[0]) != 8 or \
                    len(pieces[1]) != 4 or \
                    len(pieces[2]) != 4 or \
                    len(pieces[3]) != 4 or \
                    len(pieces[4]) != 12:
                return False
            [ int( p, 16 ) for p in pieces ]
        else:
            return False
    except ValueError: 
        return False
    except TypeError:
        return False
    return True

def to_full_uuid(uuid):
    """
    converts a short 16-bit or 32-bit reserved UUID to a full 128-bit Bluetooth
    UUID.
    """
    if not is_valid_uuid(uuid): raise ValueError("invalid UUID")
    if len(uuid) == 4:
        return "0000%s-0000-1000-8000-00805F9B34FB" % uuid
    elif len(uuid) == 8:
        return "%s-0000-1000-8000-00805F9B34FB" % uuid
    else:
        return uuid

def discover_devices(duration=8, flush_cache=True, lookup_names=False):
    """
    performs a bluetooth device discovery using the first available bluetooth
    resource.

    if lookup_names is False, returns a list of bluetooth addresses.
    if lookup_names is True, returns a list of (address, name) tuples

    duration=8 
        how long, in units of 1.28 seconds, to search for devices.  To find as
        many devices as possible, you should set this to at least 8.  
        Linux only
    flush_cache=True 
        if set to False, then discover_devices may return devices found during
        previous discoveries.
    lookup_names=False
        if set to True, then discover_devices also attempts to lookup the
        display name of each detected device.
    """
    return bt.discover_devices(flush_cache, lookup_names)

def lookup_name(address, timeout=10):
    """
    Linux only

    Tries to determine the friendly name (human readable) of the device with
    the specified bluetooth address.  Returns the name on success, and None
    on failure.

    timeout=10   how many seconds to search before giving up.
    """
    if not is_valid_address(address): 
        raise ValueError("Invalid Bluetooth address")

    return bt.lookup_name( address )

class BluetoothSocket:
    def __init__(self, proto = RFCOMM, sockfd = None):
        if proto not in [ RFCOMM ]:
            raise ValueError("invalid protocol")

        if sockfd:
            self._sockfd = sockfd
        else:
            self._sockfd = bt.socket( bt.SOCK_STREAM, bt.BTHPROTO_RFCOMM )
        self._proto = proto

    def bind(self, addrport):
        if self._proto == RFCOMM:
            addr, port = addrport

            if port == 0: port = bt.BT_PORT_ANY
            status = bt.bind( self._sockfd, addr, port )

    def listen(self, backlog):
        bt.listen( self._sockfd, backlog )

    def accept(self):
        clientfd, addr, port = bt.accept( self._sockfd )
        client = BluetoothSocket( self._proto, sockfd=clientfd )
        return client, (addr, port)

    def connect(self, addrport):
        addr, port = addrport
        bt.connect( self._sockfd, addr, port )

    def send(self, data):
        bt.send( self._sockfd, data )

    def recv(self, numbytes):
        return bt.recv( self._sockfd, numbytes )

    def close(self):
        return bt.close( self._sockfd )

    def getsockname(self):
        if self._proto == RFCOMM:
            s = bt.getsockname( self._sockfd )
            return s

    def fileno(self):
        return self._sockfd

# ============== SDP service registration and unregistration ============

def parseUUID(data, size):
    if size == 2:
        return binascii.hexlify( data )
    elif size == 4:
        return binascii.hexlify( data )
    elif size == 16:
        return "%08d-%04d-%04d-%04d-%04d%08d" % struct.unpack("!IHHHHI", data)
    else: return ValueError("invalid UUID size")

def parseInt(data, size, signed):
    fmts = { 1 : "!b" , 2 : "!h" , 4 : "!i" , 8 : "!q" , 16 : "!qq" }
    fmt = fmts[size]
    if not signed: fmt = fmt.upper()
    if fmt in [ "!qq", "!QQ" ]:
        upp, low = struct.unpack("QQ", data)
        result = ( upp << 64 ) | low
        if signed:
            result=-((~(result-1))&0x7FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFL)
        return result
    else:
        return struct.unpack(fmt, data)[0]
    
def parseTypeSizeDesc(data):
    dts = struct.unpack("B", data[0])[0]
    dtype, dsizedesc = dts >> 3, dts & 0x7
    dstart = 1
    if dsizedesc == 0:   dsize = 1
    elif dsizedesc == 1: dsize = 2
    elif dsizedesc == 2: dsize = 4
    elif dsizedesc == 3: dsize = 8
    elif dsizedesc == 4: dsize = 16
    elif dsizedesc == 5:
        dsize = struct.unpack("B", data[1])[0]
        dstart += 1
    elif dsizedesc == 6:
        dsize = struct.unpack("!H", raw[1:3])[0]
        dstart += 2
    elif dsizedesc == 7:
        dsize == struct.unpack("!I", raw[1:5])[0]
        dstart += 4

    return dtype, dsize, dstart


def parseDataElementSequence(data):
    result = []
    pos = 0
    datalen = len(data)
    while pos < datalen:
        rtype, rval, consumed = parseDataElement(data[pos:])
        pos += consumed
        result.append( (rtype, rval) )
    return result

def parseDataElement(data):
    dtype, dsize, dstart = parseTypeSizeDesc(data)
    elem = data[dstart:dstart+dsize]

    if dtype == 0:
        rtype, rval = None, None
    elif dtype == 1:
        rtype, rval = "UnsignedInt", parseInt(elem, dsize, False)
    elif dtype == 2:
        rtype, rval = "SignedInt", parseInt(elem, dsize, True)
    elif dtype == 3:
        rtype, rval = "UUID", parseUUID(elem, dsize)
    elif dtype == 4:
        rtype, rval = "String", elem
    elif dtype == 5:
        rtype, rval = "Bool", (struct.unpack("B", elem)[0] != 0)
    elif dtype == 6:
        rtype, rval = "List", parseDataElementSequence( elem )
    elif dtype == 7:
        rtype, rval = "AltList", parseDataElementSequence( elem )

    return rtype, rval, dstart+dsize

def parseRawRecord(data):
    dtype, dsize, dstart = parseTypeSizeDesc(data)
    assert dtype == 6

    pos = dstart
    datalen = len(data)
    record = {}
    while pos < datalen:
        type, attrid, consumed = parseDataElement(data[pos:])
        assert type == "UnsignedInt"
        pos += consumed
        type, attrval, consumed = parseDataElement(data[pos:])
        pos += consumed
        record[attrid] = attrval
    return record

        
def advertise_service( sock, name, service_id = "", service_classes = [], \
        profiles = [], provider = "", description = "" ):
    """
    Advertises a service with the local SDP server.  sock must be a bound,
    listening socket.  name should be the name of the service, and service_id 
    (if specified) should be a string of the form 
    "XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX", where each 'X' is a hexadecimal digit.

    service_classes is a list of service classes whose this service belongs to.
    Each class service is a 16-bit UUID in the form "XXXX", where each 'X' is a
    hexadecimal digit, or a 128-bit UUID in the form
    "XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX". There are some constants for
    standard services, e.g. SERIAL_PORT_CLASS that equals to "1101". Some class
    constants:

    SERIAL_PORT_CLASS        LAN_ACCESS_CLASS         DIALUP_NET_CLASS 
    HEADSET_CLASS            CORDLESS_TELEPHONY_CLASS AUDIO_SOURCE_CLASS
    AUDIO_SINK_CLASS         PANU_CLASS               NAP_CLASS
    GN_CLASS

    profiles is a list of service profiles that thie service fulfills. Each
    profile is a tuple with ( uuid, version ). Most standard profiles use
    standard classes as UUIDs. PyBluez offers a list of standard profiles,
    for example SERIAL_PORT_PROFILE. All standard profiles have the same
    name as the classes, except that _CLASS suffix is replaced by _PROFILE.

    provider is a text string specifying the provider of the service

    description is a text string describing the service

    A note on working with Symbian smartphones:
        bt_discover in Python for Series 60 will only detect service records
        with service class SERIAL_PORT_CLASS and profile SERIAL_PORT_PROFILE

    """
    if len(service_classes) > 0: svcclass = service_classes[0]
    else: svcclass = SERIAL_PORT_CLASS

    if len(profiles) > 0:
        sys.stderr.write("advertise_service: Ignoring service profiles\n")
    if len(service_classes) > 1:
        sys.stderr.write("advertise_service: Only using first " \
                "Service Class ID\n")
    if len(service_id) > 0:
        sys.stderr.write("advertise_service: Ignoring service class ID\n")
    if len(provider) > 0:
        sys.stderr.write("advertise_service: Ignoring service provider\n")

    svcclass = to_full_uuid( svcclass )

    bt.set_service( sock._sockfd, True, name, description, svcclass)

    sock._svcclass = svcclass
    sock._svcname = name
    sock._svcdesc = description

def stop_advertising( sock ):
    """
    Instructs the local SDP server to stop advertising the service associated
    with sock.  You should typically call this right before you close sock.
    """
    bt.set_service( sock._sockfd, False, sock._svcname, sock._svcdesc, 
            sock._svcclass)

def find_service( name = None, uuid = None, address = None ):
    """
    find_service( name = None, uuid = None, address = None )

    Searches for SDP services that match the specified criteria and returns
    the search results.  If no criteria are specified, then returns a list of
    all nearby services detected.  If more than one is specified, then
    the search results will match all the criteria specified.  If uuid is
    specified, it must be either a 16-bit UUID in the form "XXXX", where each
    'X' is a hexadecimal digit, or as a 128-bit UUID in the form
    "XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX".  A special case of address is
    "localhost", which will search for services on the local machine.

    The search results will be a list of dictionaries.  Each dictionary
    represents a search match and will have the following key/value pairs:

      host          - the bluetooth address of the device advertising the
                      service
      name          - the name of the service being advertised
      description   - a description of the service being advertised
      provider      - the name of the person/organization providing the service
      protocol      - either 'RFCOMM', 'L2CAP', None if the protocol was not
                      specified, or 'UNKNOWN' if the protocol was specified but
                      unrecognized
      port          - the L2CAP PSM # if the protocol is 'L2CAP', the RFCOMM
                      channel # if the protocol is 'RFCOMM', or None if it
                      wasn't specified
      service-classes - a list of service class IDs (UUID strings).  possibly
                        empty
      profiles        - a list of profiles - (UUID, version) pairs - the
                        service claims to support.  possibly empty.
      service-id      - the Service ID of the service.  None if it wasn't set
                        See the Bluetooth spec for the difference between
                        Service ID and Service Class ID List
    """
    if address is not None:
        addresses = [ address ]
    else:
        addresses = discover_devices(lookup_names = False)

    results = []

    for addr in addresses:
        uuidstr = uuid or PUBLIC_BROWSE_GROUP
        if not is_valid_uuid( uuidstr ): raise ValueError("invalid UUID")

        uuidstr = to_full_uuid( uuidstr )

        dresults = bt.find_service( addr, uuidstr )

        for dict in dresults:
            raw = dict["rawrecord"]

            record = parseRawRecord( raw )

            if SERVICE_CLASS_ID_LIST_ATTRID in record:
                svc_class_id_list = [ t[1] for t in \
                        record[SERVICE_CLASS_ID_LIST_ATTRID] ]
                dict["service-classes"] = svc_class_id_list
            else:
                dict["services-classes"] = []

            if BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRID in record:
                pdl = []
                for profile_desc in \
                        record[BLUETOOTH_PROFILE_DESCRIPTOR_LIST_ATTRID]:
                    uuidpair, versionpair = profile_desc[1]
                    pdl.append( (uuidpair[1], versionpair[1]) )
                dict["profiles"] = pdl
            else:
                dict["profiles"] = []

            dict["provider"] = record.get(PROVIDER_NAME_ATTRID, None)

            dict["service-id"] = record.get(SERVICE_ID_ATTRID, None)

            # XXX the C version is buggy (retrieves an extra byte or two), so
            #     get the service name here even though it may have already
            #     been set
            dict["name"] = record.get(SERVICE_NAME_ATTRID, None)

        
#        if LANGUAGE_BASE_ATTRID_LIST_ATTRID in record:
#            for triple in record[LANGUAGE_BASE_ATTRID_LIST_ATTRID]:
#                code_ISO639, encoding, base_offset = triple
#
#        if SERVICE_NAME_ATTRID in record:
#            service_name = record[SERVICE_NAME_ATTRID]
#
#        if SERVICE_DESCRIPTION_ATTRID in record:
#            service_description = record[SERVICE_DESCRIPTION_ATTRID]
            del dict["rawrecord"]

        if name is None:
            results.extend(dresults)
        else:
            results.extend( [ d for d in dresults if d["name"] == name ] )
    return results
