#!/usr/bin/env python
# Test setup script for installing bluez module.
#

from distutils.core import setup, Extension
import sys
import os

if sys.platform == 'win32':
    # XXX what's a better way to add the Platform SDK include path?
    MS_PLATFORM_SDK_PATH = \
            r"\Program Files\Microsoft Platform SDK for Windows XP SP2"

    if not os.path.exists( "%s" % MS_PLATFORM_SDK_PATH ):
        raise SystemExit("Can't find the Windows XP Platform SDK at \n%s" % \
                MS_PLATFORM_SDK_PATH)
    
    extmod = Extension( '_msbt', 
                        include_dirs = ["%s\\Include" % MS_PLATFORM_SDK_PATH],
                        library_dirs = ["%s\\Lib" % MS_PLATFORM_SDK_PATH],
                        libraries = [ "WS2_32", "Irprops" ],
                        sources=['msbt\\_msbt.c'], )
    pymods = [ 'bluetooth' ]
    package_base = "msbt"
elif sys.platform == 'linux2':
    extmod = Extension('_bluetooth',
		                 libraries = ['bluetooth'],
                         sources = ['bluez/btmodule.c', 'bluez/btsdp.c'])
    pymods = [ 'bluetooth' ]
    package_base = "bluez"

setup (	name = 'PyBluez',
       	version = '0.7',
       	description = 'python wrappers around bluez',
       	author="Albert Huang, Calvin On",
     	author_email="albert@csail.mit.edu, calvinon@mit.edu",
      	url="http://org.csail.mit.edu/pybluez",
       	ext_modules = [ extmod ],
        py_modules = pymods,
        package_dir= { '' : package_base },

# for the python cheese shop
        classifiers = [ 'Development Status :: 4 - Beta',
            'License :: OSI Approved :: GNU General Public License (GPL)',
            'Programming Language :: Python',
            'Topic :: Communications' ],
        download_url = 'http://org.csail.mit.edu/pybluez/download.html',
        long_description = 'python wrappers around bluez to allow python developers to use system bluetooth resources. PyBluez works GNU/Linux and Windows XP SP2.',
        maintainer = 'Albert Huang',
        maintainer_email = 'albert@csail.mit.edu',
        license = 'GPL',
        )
